#    This file is part of MK.
#
#    MK is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    MK is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Foobar; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import string
import os
import types
import mk

from mk import ex
from mk import env
from mk import nodes
from mk import lang

# Generate a directory name for targets, built from sources in
# directory DIR.
def _mirror_to_builddir (dir):
    if dir.startswith (env.srcdir) and not dir.startswith (env.builddir):
        path = string.replace (dir, env.srcdir, env.builddir, 1);
        if env.output_subdir:
            return '%s/%s' % (path, env.output_subdir)
        else:
            return path
    else:
        return dir

# Generate a directory name for sources, used to build targets in
# directory DIR.
def _mirror_to_srcdir (dir):
    if dir.startswith (env.builddir) and not dir.startswith (env.srcdir):
        if env.output_subdir and dir.endswith (env.output_subdir):
            dir = dir[:- len (env.output_subdir)]
            return string.replace (dir, env.builddir, env.srcdir, 1);
    else:
        return dir

# Prefix the name with the current builddir.
def _builddir_name (name):
    if not name.startswith (env.builddir):
        if env.output_subdir:
            return '%s/%s/%s' % (env.builddir, env.output_subdir, name)
        else:
            return '%s/%s' % (env.builddir, name)
    else:
        return name

# Derive an object name from a source name.  
def _object_name (name, suffix = '.o'):
    dir = os.path.dirname (name)
    base = os.path.basename (name)
    i = base.rfind ('.')
    if i != -1:
        base = base [:i]
    base = base + suffix
    return _mirror_to_builddir (dir) + '/' + base

# Derive a source name from an object name.
def _source_name (name, suffix = '.c'):
    dir = os.path.dirname (name)
    base = os.path.basename (name)
    i = base.rfind ('.')
    if i != -1:
        base = base [:i]
    base = base + suffix
    path = _mirror_to_srcdir (dir) + '/' + base
    if os.access (path, os.F_OK):
        return path
    else:
        return base

# Convert a mixed string/source node list into a list of source
# objects.
def _cook_source (src):
    if isinstance (src, types.StringTypes):
        return make_source (src)
    elif isinstance (src, nodes.source):
        return src
    else:
        raise ex.node_type_mismatch (src.name)

def _cook_sources (list):
    return [_cook_source (src) for src in list]

# Convert a mixed string/node list into a list of nodes.
def _cook_targets (list, clazz):
    targets = []
    for elt in list:
        if isinstance (elt, types.StringTypes):
            name = _builddir_name (elt)
            if not env.targets.has_key (elt):
                raise ex.missing_target (name)
            t = env.targets [name]
        else:
            t = elt
        if not isinstance (t, clazz):
            raise ex.node_type_mismatch (name)
        targets.append (t)
    return targets

# Create a phony node.  
def make_phony (name):
    if env.targets.has_key (name):
        return env.targets [name]

    return nodes.phony (name)

# Create source nodes.  
def make_source (name):
    path = mk.locate_source (name)
    if not path:
        raise ex.missing_source (name)

    if env.targets.has_key (path):
        return env.targets [path]

    return nodes.source (path)

def make_sources (lst):
    return [make_source (name) for name in lst]

# Create object file nodes.   
def make_object (obj_name, src = None, build = None):
    obj_name = _builddir_name (obj_name)
    if env.targets.has_key (obj_name):
        return env.targets [obj_name]

    if not src:
        src_name = _source_name (obj_name)
        src = make_source (src_name)
    elif isinstance (src, types.StringTypes):
        src = make_source (src)

    if not build:
        if src.lang == lang.cxx:
            build = env.cxx.compile
        else:
            build = env.cc.compile

    return nodes.object_file (obj_name, src, build)

def make_objects (objs, srcs = None, build = None):
    if srcs:
        cooked_srcs = _cook_sources (srcs)
        return [make_object (x, y, build) for x, y in zip (objs, cooked_srcs)]
    else:
        return [make_object (x, build = build) for x in objs]

def compile_source (src, obj_name = None, build = None):
    src = _cook_source (src)
    if not obj_name:
        obj_name = _object_name (src.name)
    return make_object (obj_name, src, build)

def compile_sources (srcs, obj_names = None, build = None):
    srcs = _cook_sources (srcs)
    if obj_names:
        return [compile_source (x, y, build) for x, y in zip (srcs, obj_names)]
    else:
        return [compile_source (x, build = build) for x in srcs]

def make_static_lib (name, srcs = None, objs = None, build = None):
    if env.targets.has_key (name):
        return env.targets [name]

    if srcs:
        cooked_objs = [compile_source (x) for x in _cook_sources (srcs)]
    else:
        cooked_objs = []

    if objs:
        cooked_objs.extend (objs)

    if not build:
        build = env.ar.archive
    return nodes.static_lib (_builddir_name (name), cooked_objs, build)


# Create a program target.  The ``srcs'' parameter must be a list of
# either source names or source objects.  Likewise the ``objs''
# parameter.
def _make_exec (clazz, name, srcs, objs, libs, xlibs, build):
    # Get or create the object node for each non-object elt in ``srcs''.
    if srcs:
        cooked_objs = [compile_source (x) for x in _cook_sources (srcs)]
    else:
        cooked_objs = []

    # Get object node for each non-object elt in ``objs''.
    if objs:
        cooked_objs.extend (_cook_targets (objs, nodes.object_file))

    # Get library node for each non-object elt in ``libs''.
    if libs:
        cooked_libs = \
           _cook_targets (libs, (nodes.static_lib, nodes.shared_lib))
    else:
        cooked_libs = []

    return clazz (_builddir_name (name), cooked_objs, cooked_libs,
                  xlibs, build)

def make_program (name, srcs = None, objs = None, libs = None, xlibs = None,
                  build = None):
    if not build:
        build = env.ld.link_program
    return _make_exec (nodes.program, name, srcs, objs, libs, xlibs, build)

def make_shared_lib (name, srcs = None, objs = None, libs = None,
                     xlibs = None, build = None):
    if not build:
        build = env.ld.link_shared_lib
    return _make_exec (nodes.shared_lib, name, srcs, objs, libs, xlibs, build)

def make_parser (name, gram = None):
    if not gram:
        gram = _mirror_to_srcdir (name + '.y')
    gram = _cook_source (gram)
    build = mk.tools.bison ()
    if not os.path.isabs (name):
        name = env.srcdir + '/' + name
    sname = name + '.c'
    hname = name + '.h'
    if env.targets.has_key (sname):
        src = env.targets [sname]
    else:
        src = nodes.bison_parser (sname, gram = gram, build = build.compile)
    if env.targets.has_key (hname):
        hdr = env.targets [hname]
    else:
        hdr = nodes.bison_parser (hname, gram = gram, build = build.compile)
        hdr.aux = True
    src.precious = True
    hdr.precious = True
    return src, hdr

def depends (target, source):
    target._depends  (source)

# Local variables:
# mode: python
# indent-tabs-mode: nil
# arch-tag: e859a0bd-d40a-481d-b5bc-66c217f403fb
# End:
