#    This file is part of MK.
#
#    MK is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    MK is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Foobar; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import os
import os.path
import time
import re
import string
import copy
import mk

from mk import ex
from mk import env
from mk import cache

# Debug level dependent print
def debug_print (level, fmt, tp):
    if level >= env.debug_level:
        print fmt % tp

# RE qualified search list.  
def make_match_list (dict, old = None):
    if old == None:
        result = {}
    else:
        result = old
    for x in dict.keys ():
        if result.has_key (x):
            result [x][1].extend (dict [x])
        else:
            result [x] = (re.compile (x), dict [x])
    return result

# Search for file NAME in the match list PATH.
def _locate_file (name, path, prefix = None):
    if not path:
        return None

    for txt, elt in path.iteritems ():
        if elt [0].search (name):
            for dir in elt [1]:
                if prefix:
                    pathname = '%s/%s/%s' % (prefix, dir, name)
                else:
                    pathname = '%s/%s' % (dir, name)
                if os.access (pathname, os.F_OK):
                    return os.path.abspath (pathname)
    return None

# Locate a source file NAME.  Check in order the following paths:
# a) the NAME itself
# b) the NAME prefixed with SRCDIR
# c) NAME prefixed with all matching directories in SEARCH_PATH variable
# d) paths from c) prefixed with SRCDIR
def locate_source (name):
    if os.access (name, os.F_OK):
        return os.path.abspath (name)

    path = env.srcdir + '/' + name
    if os.access (path, os.F_OK):
        return os.path.abspath (path)
        
    path = _locate_file (name, env.source_path)
    if not path:
        path = _locate_file (name, env.source_path, env.srcdir)

    return path

# Generate a relative pathname, equivalent to an absolute one.
def shortname (name):
    name = os.path.abspath (name)
    dir = os.path.dirname (name)
    cwd = os.getcwd ()

    pfx = os.path.commonprefix ([dir, cwd])
    if not pfx:
        return name

    dir = string.lstrip (dir [len (pfx):], '/')
    cwd = string.lstrip (cwd [len (pfx):], '/')

    if cwd:
        back_count = string.count (cwd, '/') + 1
    else:
        back_count = 0

    back = string.join (['..' for i in range (0, back_count)], '/')
    base = os.path.basename (name)

    if back:
        if dir:
            short = back + '/' + dir + '/' + base
        else:
            short = back + '/' + base
    elif dir:
        short = dir + '/' + base
    elif base:
        short = base
    else:
        short = '.'

    if len (short) < len (name):
        return short
    else:
        return name

# Generate a library reference, suitable for linker command line.
def libref (name):
    if name.endswith ('.a'):
        nm = name [:-2]
    else:
        i = name.rfind ('.so');
        if i != -1:
            nm = name [0:i]
        else:
            return name
    dir = os.path.dirname (nm)
    base = os.path.basename (nm)
    if base.startswith ('lib'):
        if len (dir):
            return '-L' + shortname (dir) + ' ' + '-l' + base [3:]
        else:
            return '-l' + base [3:]
    else:
        return name

# Serial number generator.
_serial = 1
def make_serial ():
    global _serial
    s = _serial
    _serial = s + 1
    return s

# Topological sort of nodes in the set.  Returns a list, where each
# element is a list of concurrently buildable nodes.
def topsort (set):
    npred = {}
    succ = {}

    for head_id in set.keys ():
        npred [head_id] = 0
        for tail in set [head_id]._auto_depend:
            if not set.has_key (tail.id):
                continue

            # Count the number of predecessors of each node
            npred [head_id] = npred [head_id] + 1

            # Associate each node with a list of successor nodes.
            if succ.has_key (tail.id):
                succ [tail.id].append (head_id)
            else:
                succ [tail.id] = [head_id]

    # Form the initial worklist of nodes with no predecessors. It's
    # also the initial batch of concurrently buildable nodes.
    worklist = filter (lambda x: npred [x] == 0, npred.keys ())
    result = [copy.copy (worklist)]

    # For each element in the current parallel batch, if it has
    # successors, decrement the predecessor count of each successor
    # and if it drops to zero, add the successor to the next parallel
    # batch.
    for id in worklist:
        next_batch = []
        if succ.has_key (id):
            for s in succ [id]:
                npred [s] = npred [s] - 1
                if npred [s] == 0:
                    next_batch.append (s)
        if next_batch:
            result.append (next_batch)
            worklist.extend (next_batch)

    return result

# Execute a mkfile, optionally recording its full path name.
def loadfile (mkfile, record = True):
    if not os.access (mkfile, os.F_OK):
        raise ex.missing_source (mkfile)
    else:
        if record:
            env.mkfiles.append (mkfile)
        execfile (mkfile, env.globals)

# Load the dependency information from a subdir makefile.
def subdir (name, makefile = 'MK'):
    old_builddir = env.builddir
    old_srcdir = env.srcdir
    try:
        env.builddir = env.builddir + '/' + name
        env.srcdir = env.srcdir + '/' + name
        makefile = '%s/%s' % (env.srcdir, makefile)
        loadfile (makefile)
    finally:
        env.builddir = old_builddir
        env.srcdir = old_srcdir

# Load the dependency information from a subdir makefile.
def subdirs (lst):
    for name in lst:
        subdir (name)

_time_offset = 0

# Get the curent time, make sure each successive invocation returns a
# different time.
def _time ():
    global _time_offset
    _time_offset = _time_offset + 0.000001
    return time.time () + _time_offset

# Run and wait for a batch of processes.  Return True if all processes
# exited with an exit status of zero or False otherwise.
def run_processes (lis):
    pids = []
    for cmd in lis:
        print ' '.join (cmd)
	pids.append (os.spawnvp (os.P_NOWAIT, cmd [0], cmd))

    success = True
    for pid in pids:
	(pid, sts) = os.waitpid (pid, 0)
	success = success and (sts == 0)
    
    return success

# Local variables:
# mode: python
# indent-tabs-mode: nil
# arch-tag: 6ddd4a15-859d-45e4-8f80-439a46c3333a
# End:
