#    This file is part of MK.
#
#    MK is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    MK is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Foobar; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import os
import types
import mk

from mk import ex
from mk import env
from mk import meta
from mk import lang

class node (object):
    __metaclass__ = meta.autoprop

    def __init__ (self, name):
        if env.targets.has_key (name):
            raise ex.target_conflict (name)
        self.id = mk.make_serial ()
        self.name = name
        self._explicit_depend = {}
        self._auto_depend = []
        self._processing = False
        self._time = 0
        self.build = None
        self.precious = False
        self.aux = False
        env.targets [name] = self

    def __getstate__ (self):
        d = self.__dict__.copy ()
        d ['_time'] = 0
        del d ['build']
        if self.build:
            d ['_build_obj'] = self.build.im_self
            d ['_build_func'] = self.build.im_func.__name__
        else:
            d ['_build_obj'] = None
            d ['_build_func'] = None
        return d

    def __setstate__ (self, state):
        self.__dict__ = state
        if self._build_obj and self._build_func:
            self.build = getattr (self._build_obj, self._build_func)
        else:
            self.build = None
        del self.__dict__ ['_build_obj']
        del self.__dict__ ['_build_func']
        env.targets [self.name] = self

    def _depends (self, dep):
        self._explicit_depend [dep.id] = dep

    def _set_time (self, tm):
        self._time = tm

    def _get_time (self):
        return self._time

    def _refresh_autodeps (self):
        # Check if we need to recompute automatic dependencies - if we
        # haven't yet done so or if some of them changed since the
        # last time we computed them.
        if not self._auto_depend:
            do_autodepend = True
        else:
            do_autodepend = False
            for nd in self._auto_depend:
                if nd._get_time () > self._autodep_time:
                    do_autodepend = True
                    break

        if not do_autodepend:
            return

        # Get the automatic dependencies build function
        if not self.build:
            autodep = None
        else:
            autodep = getattr (self.build.im_self, 'autodep', None)

        # If there's no way to build automatic dependencies just use
        # the explicit ones.
        if not autodep:
            self._auto_depend = [nd for name, nd
                                 in self._explicit_depend.iteritems ()]
        else:
            deps = {}
            for id, nd in self._explicit_depend.iteritems ():
                deps [nd.name] = True
                autodep (nd, deps);
            self._auto_depend = []
            for name, b in deps.iteritems ():
                if env.targets.has_key (name):
                    nd = env.targets [name]
                else:
                    nd = file_node (name)
                self._auto_depend.append (nd)
        self._autodep_time = mk.time ()


    def needs_update (self, set):
        if self._processing:
            print 'Circular dependency processing %s' % self.name
        else:
            # Prevent infinite recursion.
            self._processing = True

            # Refresh automatically derived dependencies.
            self._refresh_autodeps ()

            # Check if we have out-of-date or newer dependencies.
            do_update = False
            for nd in self._auto_depend:
                nd.needs_update (set)
                if nd._get_time () == 0 or nd._get_time () > self._get_time ():
                    do_update = True

            # Add to the set if there's at least one newer dependency.
            if do_update:
                set [self.id] = self
                self._set_time (0)

            self._processing = False

    def update (self):
        if self.aux:
            return None
        if self.build == None:
            print 'No rule associated with %s' % mk.shortname (self.name)
            return False
        else:
            self._set_time (mk.time ())
            return self.build (self)

    def needs_clean (self, real, set):
        if (not self.precious or real) and not set.has_key (self.id):
            set [self.id] = self
            for id, nd in self._explicit_depend.iteritems ():
                nd.needs_clean (real, set)
            self._set_time (0);

    def clean (self):
        pass

class phony (node):
    def update (self):
        return False

    def depends (self, node):
        if isinstance (node, types.ListType):
            for nd in node:
                self._depends (nd)
        else:
            self._depends (node)

class file_node (node):
    def __init__ (self, name):
        super (file_node, self).__init__ (name)
        self._need_stat = True

    def _set_time (self, tm):
        self._need_stat = False
        self._time = tm

    def _get_time (self):
        if self._need_stat:
            if os.access (self.name, os.F_OK):
                self._set_time (os.stat (self.name).st_mtime)
            else:
                self._set_time (0)
        return self._time

    def __getstate__ (self):
        d = super (file_node, self).__getstate__ ()
        d ['_need_stat'] = True
        return d

    def needs_clean (self, real, set):
        if self.build:
            super (file_node, self).needs_clean (real, set)

    def update (self):
        if not os.access (os.path.dirname (self.name), os.F_OK):
            os.makedirs (os.path.dirname (self.name))
        return super (file_node, self).update ()

    def clean (self):
        if os.access (self.name, os.F_OK):
            os.remove (self.name)

class source (file_node):
    def __init__ (self, name, lng = None):
        super (source, self).__init__ (name)
        if lng:
            self.lang = lng
        else:
            self.lang = lang.source_lang (name)

    def clean (self):
        print '*** WARNING: NOT DELETING %s ***' % self.name

class bison_parser (source):
    def __init__ (self, name, gram = None, build = None):
        super (bison_parser, self).__init__ (name)
        self.build = build
        self.precious = True
        if gram:
            self._set_grammar (gram)
        else:
            self._grammar = None

    def _set_grammar (self, gram):
        self._depends (gram)
        self._grammar = gram

    def _get_grammar (self):
        return self._grammar

    def clean (self):
        super (source, self).clean ()

class object_file (file_node):
    def __init__ (self, name, src, build):
        super (object_file, self).__init__ (name)
        if src:
            self._set_source (src)
        else:
            self._source = None
        self.build = build

    def _set_source (self, src):
        self._depends (src)
        self._source = src

    def _get_source (self):
        return self._source

    def update (self):
        if self._source == None:
            print 'No source for %s' % self.name
            return False
        else:
            return super (object_file, self).update ()

class static_lib (file_node):
    def __init__ (self, name, objs, build):
        super (static_lib, self).__init__ (name)
        if objs:
            self._set_objects (objs)
        else:
            self._objects = []
        self.build = build

    def _set_objects (self, objs):
        for obj in objs:
            self._depends (obj)
        self._objects = objs

    def _get_objects (self):
        return self._objects

    def update (self):
        if len (self._objects) == 0:
            print 'No input objects for %s' % mk.shortname (self.name)
            return False
        else:
            return super (static_lib, self).update ()

class program (file_node):
    def __init__ (self, name, objs, libs, xlibs, build):
        super (program, self).__init__ (name)
        if objs:
            self._set_objects (objs)
        else:
            self._objects = []
        if libs:
            self._set_libs (libs)
        else:
            self._libs = []
        if xlibs:
            self.xlibs = xlibs
        else:
            self.xlibs = []
        self.build = build

    def _set_objects (self, objs):
        for obj in objs:
            super (program, self)._depends (obj)
        self._objects = objs

    def _get_objects (self):
        return self._objects

    def _set_libs (self, libs):
        for lib in libs:
            super (program, self)._depends (lib)
        self._libs = libs

    def _get_libs (self):
        return self._libs

class shared_lib (program):
    def __init__ (self, name, objs, libs, xlibs, build):
        super (shared_lib, self).__init__ (name, objs, libs, xlibs, build)

# Local variables:
# mode: python
# indent-tabs-mode: nil
# arch-tag: eacf5f86-b155-4175-b23b-33dd32b5c0c4
# End:
